@@ -1,9 +1,9 @@
 package Net::DNS::SEC::DSA;
 
 #
-# $Id: DSA.pm 1619 2018-01-24 08:24:17Z willem $
+# $Id: DSA.pm 1669 2018-04-27 10:17:13Z willem $
 #
-our $VERSION = (qw$LastChangedRevision: 1619 $)[1];
+our $VERSION = (qw$LastChangedRevision: 1669 $)[1];
 
 
 =head1 NAME
@@ -29,14 +29,14 @@ generation and verification procedures.
 
     $signature = Net::DNS::SEC::DSA->sign( $sigdata, $private );
 
-Generates the wire-format binary signature from the binary sigdata
+Generates the wire-format signature from the sigdata octet string
 and the appropriate private key object.
 
 =head2 verify
 
     $validated = Net::DNS::SEC::DSA->verify( $sigdata, $keyrr, $sigbin );
 
-Verifies the signature over the binary sigdata using the specified
+Verifies the signature over the sigdata octet string using the specified
 public key resource record.
 
 =cut
@@ -44,85 +44,75 @@ public key resource record.
 use strict;
 use integer;
 use warnings;
-use Carp;
-use Digest::SHA;
 use MIME::Base64;
 
 my %DSA = (
-	3 => ['Digest::SHA'],
-	6 => ['Digest::SHA'],
+	3 => [sub { Net::DNS::SEC::libcrypto::EVP_sha1() }],
+	6 => [sub { Net::DNS::SEC::libcrypto::EVP_sha1() }],
 	);
 
 
 sub sign {
 	my ( $class, $sigdata, $private ) = @_;
 
-	my $algorithm = $private->algorithm;			# digest sigdata
-	my ( $object, @param ) = @{$DSA{$algorithm} || []};
-	die 'private key not DSA' unless $object;
-	my $hash = $object->new(@param);
-	$hash->add($sigdata);
+	my $algorithm = $private->algorithm;
+	my ($evpmd) = @{$DSA{$algorithm} || []};
+	die 'private key not DSA' unless $evpmd;
 
-	my $p = decode_base64 $private->prime;			# private key
-	my $q = decode_base64 $private->subprime;
-	my $g = decode_base64 $private->base;
-	my $x = decode_base64 $private->private_value;
-	my $y = decode_base64 $private->public_value;
+	my ( $p, $q, $g, $x, $y ) = map decode_base64( $private->$_ ),
+			qw(prime subprime base private_value public_value);
+	my $t = ( length($g) - 64 ) / 8;
 
 	my $dsa = Net::DNS::SEC::libcrypto::DSA_new();
 	Net::DNS::SEC::libcrypto::DSA_set0_pqg( $dsa, $p, $q, $g );
 	Net::DNS::SEC::libcrypto::DSA_set0_key( $dsa, $y, $x );
 
-	my $sig = Net::DNS::SEC::libcrypto::DSA_do_sign( $hash->digest, $dsa );
+	my $evpkey = Net::DNS::SEC::libcrypto::EVP_PKEY_new();
+	Net::DNS::SEC::libcrypto::EVP_PKEY_assign_DSA( $evpkey, $dsa );
 
-	$x = $q;						# destroy private key
-	Net::DNS::SEC::libcrypto::DSA_free($dsa);		# destroy private key
-
-	my $z;
-	if ($sig) {						# uncoverable branch false
-		my $t = ( length($g) - 64 ) / 8;
-		my ( $r, $s ) = Net::DNS::SEC::libcrypto::DSA_SIG_get0($sig);
-
-		# both the R and S parameters need to be 20 octets:
-		my $Rpad = 20 - length($r);
-		my $Spad = 20 - length($s);
-		$z = pack "C x$Rpad a* x$Spad a*", $t, $r, $s;	# RFC2536, section 3
-
-		Net::DNS::SEC::libcrypto::DSA_SIG_free($sig);
-	}
-	return $z;
+	my $ASN1 = Net::DNS::SEC::libcrypto::EVP_sign( $sigdata, $evpkey, &$evpmd );
+	my $l = 2 + unpack 'x3 C', $ASN1;
+	pack 'C a* a*', $t, map _BERunwrap($_), unpack "x2 a$l a*", $ASN1;
 }
 
 
 sub verify {
 	my ( $class, $sigdata, $keyrr, $sigbin ) = @_;
 
-	my $algorithm = $keyrr->algorithm;			# digest sigdata
-	my ( $object, @param ) = @{$DSA{$algorithm} || []};
-	die 'public key not DSA' unless $object;
-	my $hash = $object->new(@param);
-	$hash->add($sigdata);
-
-	my $keybin = $keyrr->keybin;				# public key
-	my $numlen = 64 + 8 * unpack( 'C', $keybin );		# RFC2536, section 2
-	my ( $q, $p, $g, $y ) = unpack "x a20 a$numlen a$numlen a$numlen", $keybin;
+	my $algorithm = $keyrr->algorithm;
+	my ($evpmd) = @{$DSA{$algorithm} || []};
+	die 'public key not DSA' unless $evpmd;
 
 	return unless $sigbin;
 
-	my $dsa = Net::DNS::SEC::libcrypto::DSA_new();
+	my $key = $keyrr->keybin;				# public key
+	my $len = 64 + 8 * unpack( 'C', $key );			# RFC2536, section 2
+	my ( $q, $p, $g, $y ) = unpack "x a20 a$len a$len a$len", $key;
 
+	my $dsa = Net::DNS::SEC::libcrypto::DSA_new();
 	Net::DNS::SEC::libcrypto::DSA_set0_pqg( $dsa, $p, $q, $g );
 	Net::DNS::SEC::libcrypto::DSA_set0_key( $dsa, $y, '' );
 
-	my ( $r, $s ) = unpack 'x a20 a20', $sigbin;		# RFC2536, section 3
-	my $dsasig = Net::DNS::SEC::libcrypto::DSA_SIG_new();
-	Net::DNS::SEC::libcrypto::DSA_SIG_set0( $dsasig, $r, $s );
+	my $evpkey = Net::DNS::SEC::libcrypto::EVP_PKEY_new();
+	Net::DNS::SEC::libcrypto::EVP_PKEY_assign_DSA( $evpkey, $dsa );
 
-	my $vrfy = Net::DNS::SEC::libcrypto::DSA_do_verify( $hash->digest, $dsasig, $dsa );
+	my ( $r, $s ) = map _BERwrap($_), unpack 'x a20 a20', $sigbin;
+	my $ASN1 = pack 'n a* a*', 0x3000 + length($r) + length($s), $r, $s;
 
-	Net::DNS::SEC::libcrypto::DSA_free($dsa);
-	Net::DNS::SEC::libcrypto::DSA_SIG_free($dsasig);
-	return $vrfy;
+	Net::DNS::SEC::libcrypto::EVP_verify( $sigdata, $ASN1, $evpkey, &$evpmd );
+}
+
+
+sub _BERwrap {
+	my $s = shift;
+	$s =~ s/^\000+(?=.)//;
+	$s =~ s/^(?=[\200-\377])/\000/;
+	pack 'C2 a*', 2, length($s), $s;
+}
+
+sub _BERunwrap {
+	my ( $l, $s ) = unpack 'x C a*', shift;
+	substr pack( "x20 a$l", $s ), -20;
 }
 
 
@@ -166,7 +156,7 @@ DEALINGS IN THE SOFTWARE.
 
 =head1 SEE ALSO
 
-L<Net::DNS>, L<Net::DNS::SEC>, L<Digest::SHA>,
+L<Net::DNS>, L<Net::DNS::SEC>,
 RFC2536,
 L<OpenSSL|http://www.openssl.org/docs>
 
